@@ -1,7 +1,10 @@
-# -*- test-case-name: twisted.conch.test.test_ssh -*-
-# Copyright (c) 2001-2007 Twisted Matrix Laboratories.
+# Copyright (c) 2001-2010 Twisted Matrix Laboratories.
 # See LICENSE for details.
 
+"""
+Tests for L{twisted.conch.ssh}.
+"""
+
 import struct
 
 try:
@@ -9,6 +12,11 @@ try:
 except ImportError:
     Crypto = None
 
+try:
+    import pyasn1
+except ImportError:
+    pyasn1 = None
+
 from twisted.conch.ssh import common, session, forwarding
 from twisted.conch import avatar, error
 from twisted.conch.test.keydata import publicRSA_openssh, privateRSA_openssh
@@ -19,7 +27,7 @@ from twisted.internet.error import ProcessTerminated
 from twisted.python import failure, log
 from twisted.trial import unittest
 
-from test_recvline import LoopbackRelay
+from twisted.conch.test.test_recvline import LoopbackRelay
 
 
 
@@ -73,7 +81,7 @@ class ConchTestAvatar(avatar.ConchUser):
         return 1
 
     def logout(self):
-        loggedOut = True
+        self.loggedOut = True
         for listener in self.listeners.values():
             log.msg('stopListening %s' % listener)
             listener.stopListening()
@@ -110,7 +118,10 @@ class ConchSessionForTestAvatar:
         self.proto = proto
         f = cmd.split()[0]
         if f == 'false':
-            FalseTransport(proto)
+            t = FalseTransport(proto)
+            # Avoid disconnecting this immediately.  If the channel is closed
+            # before execCommand even returns the caller gets confused.
+            reactor.callLater(0, t.loseConnection)
         elif f == 'echo':
             t = EchoTransport(proto)
             t.write(cmd[5:])
@@ -157,14 +168,39 @@ class CrazySubsystem(protocol.Protocol):
         good ... good
         """
 
+
+
 class FalseTransport:
+    """
+    False transport should act like a /bin/false execution, i.e. just exit with
+    nonzero status, writing nothing to the terminal.
+
+    @ivar proto: The protocol associated with this transport.
+    @ivar closed: A flag tracking whether C{loseConnection} has been called yet.
+    """
 
     def __init__(self, p):
+        """
+        @type p L{twisted.conch.ssh.session.SSHSessionProcessProtocol} instance
+        """
+        self.proto = p
         p.makeConnection(self)
-        p.processEnded(failure.Failure(ProcessTerminated(255, None, None)))
+        self.closed = 0
+
 
     def loseConnection(self):
-        pass
+        """
+        Disconnect the protocol associated with this transport.
+        """
+        if self.closed:
+            return
+        self.closed = 1
+        self.proto.inConnectionLost()
+        self.proto.outConnectionLost()
+        self.proto.errConnectionLost()
+        self.proto.processEnded(failure.Failure(ProcessTerminated(255, None, None)))
+
+
 
 class EchoTransport:
 
@@ -229,7 +265,7 @@ class SuperEchoTransport:
         self.proto.processEnded(failure.Failure(ProcessTerminated(0, None, None)))
 
 
-if Crypto: # stuff that needs PyCrypto to even import
+if Crypto is not None and pyasn1 is not None:
     from twisted.conch import checkers
     from twisted.conch.ssh import channel, connection, factory, keys
     from twisted.conch.ssh import transport, userauth
@@ -247,7 +283,7 @@ if Crypto: # stuff that needs PyCrypto to even import
     class ConchTestPublicKeyChecker(checkers.SSHPublicKeyDatabase):
         def checkKey(self, credentials):
             unittest.assertEquals(credentials.username, 'testuser', 'bad username')
-            unittest.assertEquals(credentials.blob, keys.getPublicKeyString(data=publicDSA_openssh))
+            unittest.assertEquals(credentials.blob, keys.Key.fromString(publicDSA_openssh).blob())
             return 1
 
     class ConchTestPasswordChecker:
@@ -288,14 +324,14 @@ if Crypto: # stuff that needs PyCrypto to even import
 
         def getPublicKeys(self):
             return {
-                'ssh-rsa':keys.getPublicKeyString(data=publicRSA_openssh),
-                'ssh-dss':keys.getPublicKeyString(data=publicDSA_openssh)
+                'ssh-rsa': keys.Key.fromString(publicRSA_openssh),
+                'ssh-dss': keys.Key.fromString(publicDSA_openssh)
             }
 
         def getPrivateKeys(self):
             return {
-                'ssh-rsa':keys.getPrivateKeyObject(data=privateRSA_openssh),
-                'ssh-dss':keys.getPrivateKeyObject(data=privateDSA_openssh)
+                'ssh-rsa': keys.Key.fromString(privateRSA_openssh),
+                'ssh-dss': keys.Key.fromString(privateDSA_openssh)
             }
 
         def getPrimes(self):
@@ -309,7 +345,6 @@ if Crypto: # stuff that needs PyCrypto to even import
     class ConchTestBase:
 
         done = 0
-        allowedToError = 0
 
         def connectionLost(self, reason):
             if self.done:
@@ -320,9 +355,15 @@ if Crypto: # stuff that needs PyCrypto to even import
 
         def receiveError(self, reasonCode, desc):
             self.expectedLoseConnection = 1
-            if not self.allowedToError:
-                unittest.fail('got disconnect for %s: reason %s, desc: %s' %
-                               (self, reasonCode, desc))
+            # Some versions of OpenSSH (for example, OpenSSH_5.3p1) will
+            # send a DISCONNECT_BY_APPLICATION error before closing the
+            # connection.  Other, older versions (for example,
+            # OpenSSH_5.1p1), won't.  So accept this particular error here,
+            # but no others.
+            unittest.assertEquals(
+                reasonCode, transport.DISCONNECT_BY_APPLICATION,
+                'got disconnect for %s: reason %s, desc: %s' % (
+                    self, reasonCode, desc))
             self.loseConnection()
 
         def receiveUnimplemented(self, seqID):
@@ -337,19 +378,25 @@ if Crypto: # stuff that needs PyCrypto to even import
             transport.SSHServerTransport.connectionLost(self, reason)
 
     class ConchTestClient(ConchTestBase, transport.SSHClientTransport):
+        """
+        @ivar completed: A L{Deferred} which will fire when the expected number
+            of results have been collected.
+        """
+        def __init__(self):
+            self.completed = defer.Deferred()
 
         def connectionLost(self, reason):
             ConchTestBase.connectionLost(self, reason)
             transport.SSHClientTransport.connectionLost(self, reason)
 
         def verifyHostKey(self, key, fp):
-            unittest.assertEquals(key, keys.getPublicKeyString(data = publicRSA_openssh))
+            unittest.assertEquals(key, keys.Key.fromString(publicRSA_openssh).blob())
             unittest.assertEquals(fp,'3d:13:5f:cb:c9:79:8a:93:06:27:65:bc:3d:0b:8f:af')
             return defer.succeed(1)
 
         def connectionSecure(self):
             self.requestService(ConchTestClientAuth('testuser',
-                ConchTestClientConnection()))
+                ConchTestClientConnection(self.completed)))
 
     class ConchTestClientAuth(userauth.SSHUserAuthClient):
 
@@ -368,17 +415,25 @@ if Crypto: # stuff that needs PyCrypto to even import
 
         def getPrivateKey(self):
             self.canSucceedPublicKey = 1
-            return defer.succeed(keys.getPrivateKeyObject(data=privateDSA_openssh))
+            return defer.succeed(keys.Key.fromString(privateDSA_openssh))
 
         def getPublicKey(self):
-            return keys.getPublicKeyString(data=publicDSA_openssh)
+            return keys.Key.fromString(publicDSA_openssh)
 
     class ConchTestClientConnection(connection.SSHConnection):
-
+        """
+        @ivar _completed: A L{Deferred} which will be fired when the number of
+            results collected reaches C{totalResults}.
+        """
         name = 'ssh-connection'
         results = 0
         totalResults = 8
 
+        def __init__(self, completed):
+            connection.SSHConnection.__init__(self)
+            self._completed = completed
+
+
         def serviceStarted(self):
             self.openChannel(SSHTestFailExecChannel(conn = self))
             self.openChannel(SSHTestFalseChannel(conn = self))
@@ -395,6 +450,7 @@ if Crypto: # stuff that needs PyCrypto to even import
             if self.results == self.totalResults:
                 self.transport.expectedLoseConnection = 1
                 self.serviceStopped()
+                self._completed.callback(None)
 
     class SSHUnknownChannel(channel.SSHChannel):
 
@@ -405,7 +461,6 @@ if Crypto: # stuff that needs PyCrypto to even import
             good .... good
             """
             log.msg('unknown open failed')
-            log.flushErrors()
             self.conn.addResult()
 
         def channelOpen(self, ignored):
@@ -429,7 +484,6 @@ if Crypto: # stuff that needs PyCrypto to even import
 
         def _ebRequestWorked(self, ignored):
             log.msg('fail exec finished')
-            log.flushErrors()
             self.conn.addResult()
             self.loseConnection()
 
@@ -446,23 +500,39 @@ if Crypto: # stuff that needs PyCrypto to even import
             d.addErrback(self._ebRequestFailed)
             log.msg('opened false')
 
+
         def _cbRequestWorked(self, ignored):
-            pass
+            """
+            The false channel request is expected to succeed. No need to check
+            for this to be actually called, since it will automatically be
+            errbacked if this channel is closed with the deferred left pending.
+            """
+
 
         def _ebRequestFailed(self, reason):
-            unittest.fail('false exec failed: %s' % reason)
+            """
+            The request deferred should never errback.
+            """
+            log.err(Exception('False channel request errbacked.'))
+
 
         def dataReceived(self, data):
             unittest.fail('got data when using false')
 
+
         def request_exit_status(self, status):
-            status, = struct.unpack('>L', status)
-            if status == 0:
-                unittest.fail('false exit status was 0')
+            self.status, = struct.unpack('>L', status)
+
+
+        def closed(self):
+            if self.status == 0:
+                log.err(Exception('false exit status was 0'))
             log.msg('finished false')
             self.conn.addResult()
             return 1
 
+
+
     class SSHTestEchoChannel(channel.SSHChannel):
 
         name = 'session'
@@ -691,8 +761,14 @@ class SSHProtocolTestCase(unittest.TestCase):
     if not Crypto:
         skip = "can't run w/o PyCrypto"
 
-    def testOurServerOurClient(self):
-        """test the Conch server against the Conch client
+    if not pyasn1:
+        skip = "can't run w/o PyASN1"
+
+    def test_ourServerOurClient(self):
+        """
+        Run the Conch server against the Conch client.  Set up several different
+        channels which exercise different behaviors and wait for them to
+        complete.  Verify that the channels with errors log them.
         """
         realm = ConchTestRealm()
         p = portal.Portal(realm)
@@ -716,7 +792,28 @@ class SSHProtocolTestCase(unittest.TestCase):
                                 self.serverTransport.clearBuffer)
             log.callWithContext({'system': 'clientTransport'},
                                 self.clientTransport.clearBuffer)
-        self.failIf(self.server.done and self.client.done)
+        self.assertFalse(self.server.done and self.client.done)
+
+        return self.client.completed.addCallback(self._verifyExpectedErrors)
+
+    def _verifyExpectedErrors(self, ignored):
+        """
+        Verify errors of the current test case are as many as we expect and the
+        ones we expect.
+        """
+        errors = self.flushLoggedErrors(error.ConchError)
+        errors.sort(key=lambda reason: reason.value.args)
+
+        # Two errors exactly are expected.  Report the whole list if we get a
+        # different number.
+        self.assertEquals(
+            len(errors), 2, "Expected two errors, got: %r" % (errors,))
+
+        # SSHUnknownChannel causes this to be logged.
+        self.assertEquals(errors[0].value.args, (3, 'unknown channel'))
+        # SSHTestFailExecChannel causes this to be logged.
+        self.assertEquals(errors[1].value.args, ('bad exec', None))
+
 
 
 class TestSSHFactory(unittest.TestCase):
@@ -724,6 +821,9 @@ class TestSSHFactory(unittest.TestCase):
     if not Crypto:
         skip = "can't run w/o PyCrypto"
 
+    if not pyasn1:
+        skip = "can't run w/o PyASN1"
+
     def makeSSHFactory(self, primes=None):
         sshFactory = factory.SSHFactory()
         gpk = lambda: {'ssh-rsa' : keys.Key(None)}
@@ -754,7 +854,7 @@ class TestSSHFactory(unittest.TestCase):
             return transport.SSHServerTransport()
         factory = self.makeSSHFactory()
         factory.protocol = makeProtocol
-        protocol = factory.buildProtocol(None)
+        factory.buildProtocol(None)
         self.assertEquals([()], calls)
 
 
@@ -769,23 +869,6 @@ class TestSSHFactory(unittest.TestCase):
                 p2.supportedKeyExchanges)
 
 
-class EntropyTestCase(unittest.TestCase):
-    """
-    Tests for L{common.entropy}.
-    """
-
-    def test_deprecation(self):
-        """
-        Test the deprecation of L{common.entropy.get_bytes}.
-        """
-        def wrapper():
-            return common.entropy.get_bytes(10)
-        self.assertWarns(DeprecationWarning,
-            "entropy.get_bytes is deprecated, please use "
-            "twisted.python.randbytes.secureRandom instead.",
-            __file__, wrapper)
-
-
 
 class MPTestCase(unittest.TestCase):
     """
@@ -799,6 +882,9 @@ class MPTestCase(unittest.TestCase):
     if not Crypto:
         skip = "can't run w/o PyCrypto"
 
+    if not pyasn1:
+        skip = "can't run w/o PyASN1"
+
 
     def test_getMP(self):
         """
